"""
References
---------
1. PyNite: An open-source FEA solver for python; https://github.com/JWock82/PyNite
2. Bathe, K. J. (2006). Finite element procedures. Klaus-Jurgen Bathe.
"""

#Jax and standard numpy
from jax import jit,vmap
import jax.numpy as jnp
import numpy as np
from jax.experimental import sparse #sparse module of Jax, under active development

#Partial
from functools import partial

#JaxSSO
from .Node import Node #'Node' objects
from .BeamCol import BeamCol #'BeamCol' objects
from . import BeamCol as BeamColSens # Import BeamCol.py module 


class Model_Sens():
    """
    Objects for sensitivity analysis model in JaxSSO
    Including:
    1. Defining the model that is necessary for global stiffness matrix (adding nodes, elements)
    2. Assemble the global stiffness matrix(K) without considering boundary conditions
    3. Get the sensitivity of structure's global stiffness matrix 'K' w.r.t. nodal coordinates
    Not including:
    1. Assigning loads
    2. Assigning boundary conditions
    3. Solving [K]{u} = {f}
    """

    def __init__(self):
        """
        Initialize a 'Model_Sens' object.
        Create lists storing:
            1. Nodes
            2. BeamCol elements
        
        """
        
        self.nodes = {}      # A dictionary of the structure's nodes
        self.beamcols = {}    # A dictionary of the structure's beam-columns
        self.supports_indices = [] # A list storing all the indices of the active supports 



    def node(self, nodeTag, X, Y, Z):
        """
        Adding/updating a node to the model.

        Inputs:
        -----
        nodeTag: int 
            the tag/index of this node

        X, Y, Z: float
            the coordinates of thie node

        """
        
        # Create a new node
        new_node = Node(nodeTag, X, Y, Z)
        
        # Add the new node/updating an existing node to model\
        self.nodes[nodeTag] = new_node

    def beamcol(self, eleTag, i_node, j_node, E, G, Iy, Iz, J, A):
        '''
        Adding/updating a beam-column element to the model
        
        Inputs
        ----------
        eleTag : int
            Index of this element

        i_node : int
            The tag of the i-node (start node).

        j_node : int
            The tag of the j-node (end node).

        E: float
            Young's modulus

        G: float
            Shear modulus

        Iy,Iz: float
            Moment of inertia about local y and local z axis

        J: float
            Torsional moment of inertia of cross section

        A: float
            Area of cross section
        '''
        # Nodes' coordinates
        x1,y1,z1 = [self.nodes[i_node].X, self.nodes[i_node].Y, self.nodes[i_node].Z]
        x2,y2,z2 = [self.nodes[j_node].X, self.nodes[j_node].Y, self.nodes[j_node].Z]

        #Create/update a beam-column and save it to the dictionary
        new_beamcol = BeamCol(eleTag, i_node, j_node, x1, y1, z1, x2, y2, z2, E, G, Iy, Iz, J, A)
        self.beamcols[eleTag] = new_beamcol


    def add_support(self,node_index,active_supports=[1,1,1,1,1,1]):
        '''
        Adding nodal support in the model.
        Inputs
        -----
        node_index: int
            The index of the node.
        
        active_supports: list, 1darray
            Whether the support for [X,Y,Z,RX,RY,RZ] is active.
            1 is active, 0 is deactivated.
            Default is fixed.
        '''
        indices = np.linspace(node_index*6,node_index*6+5,6,dtype=int)
        active_supports = np.array(active_supports,dtype=int) #convert to standard np.array
        where_active = np.argwhere(active_supports==1).ravel() #which dofs are active
        self.supports_indices.append(indices[where_active]) #append to the global list

    def add_nodal_load(self,f):
        '''
        Add the global load vector {f}.
        Inputs
        ---
        f: ndarray of shape (6*n_node)
            The global nodal loading vector.
        '''
        self.f = f 
        
    def K(self):
        '''
        Return the global stiffness of the system in 'jax.experimental.sparce.BCOO' format.
        Without the boundary conditions assigned.
        
        '''
        n_dof = 6*len(self.nodes) #number of dof
        if self.beamcols:
            #Creating containers storing attributes
            beamcol_list = list(self.beamcols.values())  #create a list for the beamcols

            beamcol_eleTag = np.array([beamcol.eleTag for beamcol in beamcol_list])  #Tags of beamcols
            beamcol_i_nodeTag = np.array([beamcol.i_nodeTag for beamcol in beamcol_list])  #Tags of i-node of beamcols
            beamcol_j_nodeTag = np.array([beamcol.j_nodeTag for beamcol in beamcol_list])  #Tags of j-node of beamcols
            beamcol_x1 = np.array([beamcol.x1 for beamcol in beamcol_list],dtype=float) #x of inode of beamcols
            beamcol_y1 = np.array([beamcol.y1 for beamcol in beamcol_list],dtype=float) #y of inode of beamcols
            beamcol_z1 = np.array([beamcol.z1 for beamcol in beamcol_list],dtype=float) #z of inode of beamcols
            beamcol_x2 = np.array([beamcol.x2 for beamcol in beamcol_list],dtype=float) #x of jnode of beamcols
            beamcol_y2 = np.array([beamcol.y2 for beamcol in beamcol_list],dtype=float) #y of jnode of beamcols
            beamcol_z2 = np.array([beamcol.z2 for beamcol in beamcol_list],dtype=float) #z of jnode of beamcols
            beamcol_E = np.array([beamcol.E for beamcol in beamcol_list],dtype=float) #E of beamcols
            beamcol_G = np.array([beamcol.G for beamcol in beamcol_list],dtype=float) #G of beamcols
            beamcol_Iy = np.array([beamcol.Iy for beamcol in beamcol_list],dtype=float) #Iy of beamcols
            beamcol_Iz = np.array([beamcol.Iz for beamcol in beamcol_list],dtype=float) #Iy of beamcols
            beamcol_J = np.array([beamcol.J for beamcol in beamcol_list],dtype=float) #J of beamcols
            beamcol_A = np.array([beamcol.A for beamcol in beamcol_list],dtype=float) #A of beamcols

            #Create 'BeamCol' object from arrays of attributes
            beamcols_replica = BeamCol(beamcol_eleTag, beamcol_i_nodeTag, beamcol_j_nodeTag, 
                                                beamcol_x1,beamcol_y1,beamcol_z1,beamcol_x2,beamcol_y2,
                                                beamcol_z2,beamcol_E,beamcol_G,beamcol_Iy,beamcol_Iz,beamcol_J,beamcol_A)

            # Call the functions defined for 'BeamCol' objects
            # Implement jax.vmap
            data,indices = vmap(BeamColSens.Ele_K_to_Global)(beamcols_replica)  #Get the indices and values storing all the local stiffness matrices
            data = data.reshape(-1) # re-dimension the data to 1darray (144*n_beamcol,)
            indices = indices.reshape(-1,2) #re-dimension to 2darray of shape (144*n_beamcol,2)
            K_global = sparse.BCOO((data,indices),shape=(n_dof,n_dof))
        
        return K_global

    def solve(self):
        '''
        Solving [K]{u} = {f}
        '''
        K_dense = self.K().todense() #Convert to dense matrix
        known_indices,unknown_indices = self.bc_indices() #Get the bc indices

        #Parition the global K into four blocks
        #based on the boundary conditions
        K11 = K_dense[unknown_indices,:][:,unknown_indices] #K11
        #K12 = K_dense[unknown_indices,:][:,known_indices]
        #K21 = K_dense[known_indices,:][:,unknown_indices]
        #K22 = K_dense[known_indices,:][:,known_indices]

        #Load vector @ unknown indices
        f_uk = self.f[unknown_indices]

        #Solving FEA
        u_unknown = jnp.linalg.solve(K11,f_uk)

        #Global displacement vector
        u_all = jnp.zeros(6*len(self.nodes))
        u_all = u_all.at[unknown_indices].set(u_unknown)

        self.u = u_all #store the displacement vector

    def bc_indices(self):
        '''
        Return the indices of unknown and known dof-displacement based on the boundary conditions.
        Returns
        ------
        known_indices: ndarray 
            indices of displacement that are known. 
        
        unknown_indices: ndarray
            indices of unknown displacement
        '''
        known_indices = (np.array(self.supports_indices,dtype=int)).ravel() #convert the active support indices list to np.array
        all_indices = np.linspace(0,6*len(self.nodes)-1,6*len(self.nodes),dtype=int) #create a container for all indices
        unknown_indices = all_indices[np.where(np.isin(all_indices,known_indices,assume_unique=True,invert=True))] #slice out the known indices 
        return known_indices,unknown_indices

    def Sens_C_Coord(self,u):
        '''
        Return the sensitivity of the strain energy (compliance) of the system wrt the displacement vector 'u'.

        Inputs
        -----
        u: ndarray of shape (6*n_node)
            Displacement of each DOF

        Returns
        -----
        dcdx: ndarray of shape(3*n_node)
            Gradient of the strain energy wrt nodal coordinates
        '''
        dcdx_all = jnp.zeros(3*len(self.nodes)) #container storing the sensitivity
        if self.beamcols:
            
            #Creating containers storing attributes
            beamcol_list = list(self.beamcols.values())  #create a list for the beamcols

            beamcol_eleTag = np.array([beamcol.eleTag for beamcol in beamcol_list])  #Tags of beamcols
            beamcol_i_nodeTag = np.array([beamcol.i_nodeTag for beamcol in beamcol_list])  #Tags of i-node of beamcols
            beamcol_j_nodeTag = np.array([beamcol.j_nodeTag for beamcol in beamcol_list])  #Tags of j-node of beamcols
            beamcol_x1 = np.array([beamcol.x1 for beamcol in beamcol_list],dtype=float) #x of inode of beamcols
            beamcol_y1 = np.array([beamcol.y1 for beamcol in beamcol_list],dtype=float) #y of inode of beamcols
            beamcol_z1 = np.array([beamcol.z1 for beamcol in beamcol_list],dtype=float) #z of inode of beamcols
            beamcol_x2 = np.array([beamcol.x2 for beamcol in beamcol_list],dtype=float) #x of jnode of beamcols
            beamcol_y2 = np.array([beamcol.y2 for beamcol in beamcol_list],dtype=float) #y of jnode of beamcols
            beamcol_z2 = np.array([beamcol.z2 for beamcol in beamcol_list],dtype=float) #z of jnode of beamcols
            beamcol_E = np.array([beamcol.E for beamcol in beamcol_list],dtype=float) #E of beamcols
            beamcol_G = np.array([beamcol.G for beamcol in beamcol_list],dtype=float) #G of beamcols
            beamcol_Iy = np.array([beamcol.Iy for beamcol in beamcol_list],dtype=float) #Iy of beamcols
            beamcol_Iz = np.array([beamcol.Iz for beamcol in beamcol_list],dtype=float) #Iy of beamcols
            beamcol_J = np.array([beamcol.J for beamcol in beamcol_list],dtype=float) #J of beamcols
            beamcol_A = np.array([beamcol.A for beamcol in beamcol_list],dtype=float) #A of beamcols

            #Create 'BeamCol' object from arrays of attributes
            beamcols_replica = BeamCol(beamcol_eleTag, beamcol_i_nodeTag, beamcol_j_nodeTag, 
                                                beamcol_x1,beamcol_y1,beamcol_z1,beamcol_x2,beamcol_y2,
                                                beamcol_z2,beamcol_E,beamcol_G,beamcol_Iy,beamcol_Iz,beamcol_J,beamcol_A)

            # Call the functions defined for 'BeamCol' objects
            # Implement jax.vmap and jax.jit to boost the calculation
            # Note that the first call of jax.jit usually takes long because it is "compiling" the codes for future fast runs
            # the following calls will be extremely fast
            
            beamcol_SensKCoord = jnp.array(vmap(BeamColSens.Ele_Sens_K_Coord)(beamcols_replica))  #Get the sensitivity, shape of (6,n_beamcol,12,12)
            
            #From dkdx to dcdx
            dcdx_bc = dcdx_beamcol(beamcol_i_nodeTag,beamcol_j_nodeTag,beamcol_SensKCoord,u)
            
            #Add to the global array
            dcdx_all += dcdx_bc

        return dcdx_all


#External functions

def dcdx_beamcol_expanded(i,j,dkdx,u):
    '''
    Return an ndarray of shape (n_beamcol,3*n_node), representing the sensitivity of the 
    strain energy contributed by each beam-column.
    Each row represents the dcdx contribution from each beamcolumn.
    This function is now written for one beam-column but will be 'vmapped' later to be applied to
    all the beam columns in the system.
    The following inputs and returns are for the 'vmapped' function.

    Inputs
    -----
    i: ndarray of shape (n_beamcol)
        i-node of each beamcolumn,

    j: ndarray of shape (n_beamcol)
        j-node of each beamcolumn

    dkdx: ndarray of shape (6,n_beamcol,12,12)
        sensitivity of local stiffness wrt to nodal coordinates of each beam-column

    u: ndarray of shape (6*n_node)
        displacement vector in global coordinate system

    Returns
    -----
    dcdx_g: ndarray of shape (n_beamcol,3*n_node)
        sensitivity of the strain energy contributed by each beam-column.

    '''

    index_i_node = jnp.linspace(i*6,i*6+5,6,dtype=int) #index of i-node
    index_j_node = jnp.linspace(j*6,j*6+5,6,dtype=int) #index of j-node
    index_beamcol = jnp.hstack((index_i_node,index_j_node)) #stack 'em
    u_e = jnp.asarray(u,dtype=float)[index_beamcol] #displacement vector of this beamcolumn
    dcdx_e = -0.5*u_e.T@dkdx@u_e #adjoint method for sensitivity
    n_node = u.shape[0]/6
    dcdx_g = jnp.zeros(int(3*n_node)) #extened container
    index_i_crd = jnp.linspace(i*3,i*3+2,3,dtype=int) #index for coordinate
    index_j_crd = jnp.linspace(j*3,j*3+2,3,dtype=int) #index for coordiante
    index_crd = jnp.hstack((index_i_crd,index_j_crd)) #stack 'em
    dcdx_g = dcdx_g.at[index_crd].set(dcdx_e) #update the array
    return dcdx_g

@jit
def dcdx_beamcol(i_s,j_s,dkdx,u):
    '''
    Sensitivity of the strain energy wrt nodal coordinates contributed by beam-column elements.

    Inputs
    -----
    i_s: ndarray of shape (n_beamcol)
        i-node of each beamcolumn,

    j_s: ndarray of shape (n_beamcol)
        j-node of each beamcolumn

    dkdx: ndarray of shape (6,n_beamcol,12,12)
        sensitivity of local stiffness wrt to nodal coordinates of each beam-column

    u: ndarray of shape (6*n_node)
        displacement vector in global coordinate system

    Returns
    -----
    ndarray of shape (3*n_node)
        sensitivity of the strain energy wrt nodal coordinates contributed by beam-column.
    '''
    dcdx_ex = vmap(dcdx_beamcol_expanded,(0,0,1,None),0)(i_s,j_s,dkdx,u)
    return jnp.sum(dcdx_ex,axis=0) 


"""
References
---------
1. PyNite: An open-source FEA solver for python; https://github.com/JWock82/PyNite
2. Bathe, K. J. (2006). Finite element procedures. Klaus-Jurgen Bathe.
"""

#Import packages needed
from jax.numpy import array, zeros, matmul, sqrt
from jax.numpy.linalg import solve,norm
import jax.numpy as jnp
from jax import jacfwd,jit
from jax.tree_util import register_pytree_node_class
import numpy as np




@register_pytree_node_class  #Register 'BeamCol' objects as a valid jax type
class BeamCol():
    '''
    A class for an elastic beam-column element.
    We register this class to JAX, so that it is a valid JAX type class.

    Inputs
    -----
    eleTag: int
        The tag of this element

    i_nodeTag, j_nodeTag: int
        Node tag of i-node and j-node of the element

    x1,y1,z1,x2,y2,z2: float
        The coordinate of the i-th node and the j-th node of this element (xi,yi,zi,xj,yj,zj).

    E: float
        Young's modulus

    G: float
        Shear modulus

    Iy,Iz: float
        Moment of inertia about local y and local z axis

    J: float
        Torsional moment of inertia of cross section

    A: float
        Area of cross section
    '''
        
    def __init__(self, eleTag, i_nodeTag, j_nodeTag, x1, y1, z1, x2, y2, z2, E, G, Iy, Iz, J, A):
        
        #Inputs, Children of Pytree
        self.eleTag = eleTag    # Element tag
        self.i_nodeTag = i_nodeTag #i-node's tag
        self.j_nodeTag = j_nodeTag #j-node's tag
        self.x1,self.y1,self.z1,self.x2,self.y2,self.z2 = [x1,y1,z1,x2,y2,z2] #Coordinates
        self.E = E  # The modulus of elasticity of the element
        self.G = G  # The shear modulus of the element
        self.Iy = Iy  # The y-axis moment of inertia
        self.Iz = Iz  # The z-axis moment of inertia
        self.J = J  # The torsional constant
        self.A = A  # The cross-sectional area
                

    def L(self):
        '''
        Return the length of the beam-column
        '''
        return sqrt((self.x1 - self.x2)**2+(self.y1-self.y2)**2+(self.z1 - self.z2)**2)

    def T(self):
        '''
        Returns the transformation matrix between local and global axis.
        '''
        x1 = self.x1
        x2 = self.x2
        y1 = self.y1
        y2 = self.y2
        z1 = self.z1
        z2 = self.z2
        zero_entry = 0
        one_entry = 1
        C = array([(x2-x1)/self.L(), (y2-y1)/self.L(), (z2-z1)/self.L()]) #unit vector
    

        #Projections of C 
        Cxz_vec = array([(x2-x1)/self.L(), zero_entry, (z2-z1)/self.L()]) 
        Cxz = norm(Cxz_vec,axis=0)
        Cx = (x2-x1)/self.L() #projection on x
        Cy = (y2-y1)/self.L() #projection on y
        Cz = (z2-z1)/self.L() #projection on z
        sin_theta = Cxz
        cos_theta = Cy
        sin_alpha = zero_entry
        cos_alpha = one_entry

        #Check if the member only has y-difference
        dirCos = jnp.where(Cxz==0,jnp.array([[zero_entry, Cy, zero_entry],
                           [-Cy*cos_alpha, zero_entry, sin_alpha],
                           [Cy*cos_alpha,  zero_entry, cos_alpha]]),
                      jnp.array([[Cx,Cy,Cz],
                            [(-Cx*Cy*cos_alpha-Cz*sin_alpha)/Cxz, Cxz*cos_alpha,(-Cy*Cz*cos_alpha+Cx*sin_alpha)/Cxz],
                                 [(Cx*Cy*sin_alpha-Cz*cos_alpha)/Cxz, -Cxz*sin_alpha, (Cy*Cz*sin_alpha+Cx*cos_alpha)/Cxz ]]))

        # Build the transformation matrix
        # !!!!Should be optimized in the future, less operations with .at.set
        transMatrix = zeros((12, 12))
        transMatrix = transMatrix.at[0:3,0:3].set(dirCos)
        transMatrix = transMatrix.at[3:6, 3:6].set(dirCos)
        transMatrix = transMatrix.at[6:9, 6:9].set(dirCos)
        transMatrix = transMatrix.at[9:12, 9:12].set(dirCos)
        return transMatrix


    def K_local(self):
        '''
        Return the element's stiffness matrix WITHOUT coordinate transformation
        '''
        E = self.E
        G = self.G
        Iy = self.Iy
        Iz = self.Iz
        J = self.J
        A = self.A
        L = self.L()
        zeros_entry = 0 #zero entry in the matrix
        k = jnp.array([[A*E/L,  zeros_entry,             zeros_entry,             zeros_entry,      zeros_entry,            zeros_entry,            -A*E/L, zeros_entry,             zeros_entry,             zeros_entry,      zeros_entry,            zeros_entry],
                        [zeros_entry,      12*E*Iz/L**3,  zeros_entry,             zeros_entry,      zeros_entry,            6*E*Iz/L**2,  zeros_entry,      -12*E*Iz/L**3, zeros_entry,             zeros_entry,      zeros_entry,            6*E*Iz/L**2],
                        [zeros_entry,      zeros_entry,             12*E*Iy/L**3,  zeros_entry,      -6*E*Iy/L**2, zeros_entry,            zeros_entry,      zeros_entry,             -12*E*Iy/L**3, zeros_entry,      -6*E*Iy/L**2, zeros_entry],
                        [zeros_entry,      zeros_entry,             zeros_entry,             G*J/L,  zeros_entry,            zeros_entry,            zeros_entry,      zeros_entry,             zeros_entry,             -G*J/L, zeros_entry,            zeros_entry],
                        [zeros_entry,      zeros_entry,             -6*E*Iy/L**2,  zeros_entry,      4*E*Iy/L,     zeros_entry,            zeros_entry,      zeros_entry,             6*E*Iy/L**2,   zeros_entry,      2*E*Iy/L,     zeros_entry],
                        [zeros_entry,      6*E*Iz/L**2,   zeros_entry,             zeros_entry,      zeros_entry,            4*E*Iz/L,     zeros_entry,      -6*E*Iz/L**2,  zeros_entry,             zeros_entry,      zeros_entry,            2*E*Iz/L],
                        [-A*E/L, zeros_entry,             zeros_entry,             zeros_entry,      zeros_entry,            zeros_entry,            A*E/L,  zeros_entry,             zeros_entry,             zeros_entry,      zeros_entry,            zeros_entry],
                        [zeros_entry,      -12*E*Iz/L**3, zeros_entry,             zeros_entry,      zeros_entry,            -6*E*Iz/L**2, zeros_entry,      12*E*Iz/L**3,  zeros_entry,             zeros_entry,      zeros_entry,            -6*E*Iz/L**2],
                        [zeros_entry,      zeros_entry,             -12*E*Iy/L**3, zeros_entry,      6*E*Iy/L**2,  zeros_entry,            zeros_entry,      zeros_entry,             12*E*Iy/L**3,  zeros_entry,      6*E*Iy/L**2,  zeros_entry],
                        [zeros_entry,      zeros_entry,             zeros_entry,             -G*J/L, zeros_entry,            zeros_entry,            zeros_entry,      zeros_entry,             zeros_entry,             G*J/L,  zeros_entry,            zeros_entry],
                        [zeros_entry,      zeros_entry,             -6*E*Iy/L**2,  zeros_entry,      2*E*Iy/L,     zeros_entry,            zeros_entry,      zeros_entry,             6*E*Iy/L**2,   zeros_entry,      4*E*Iy/L,     zeros_entry],
                        [zeros_entry,      6*E*Iz/L**2,   zeros_entry,             zeros_entry,      zeros_entry,            2*E*Iz/L,     zeros_entry,      -6*E*Iz/L**2,  zeros_entry,             zeros_entry,      zeros_entry,            4*E*Iz/L]])
    
        return k
    
    def K(self):
        '''
        Return the element's stiffness matrix in global coordinate system

        '''
        return matmul(solve(self.T(),self.K_local()),self.T())
    

    #Tell JAX how to flatten this class
    def tree_flatten(self):
        children = (self.eleTag,self.i_nodeTag, self.j_nodeTag, self.x1,self.y1,self.z1,self.x2,self.y2,self.z2,self.E,self.G,self.Iy,self.Iz,self.J,self.A)
        aux_data = None
        return (children, aux_data)

    #Tell JAX how to unflatten this class
    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(*children)    



'''
The following functions calculate the sensitivity of element's stiffness matrix K w.r.t its coordinates.

In order to implement Jax's 'Vmap' functionality later to boost the calculation of sensitivity, 
Note that functions are defined for one 'BeamCol' object, not a list of 'BeamCol' objects, this is because
'jax.vmap' works for objects(Python.containers), not containers of objects. 

These functions are:
    ----
    BeamCol_K(eleTag, x1, y1, z1, x2, y2, z2,E,G,Iy,Iz,J,A):
        Create a 'BeamCol' object and return the global stiffness matrix of this element
    ----
    
    ----
    Ele_Sens_K_Coord(BeamCol):
        Implement Jax.jacfwd to calculate the Jacobian.
        Return the sensitivity of the global stiffness of this member w.r.t. the coordinates of the member.
    ----

'''


#Stiffness matrix of the element
def BeamCol_K(eleTag, i_nodeTag, j_nodeTag, x1, y1, z1, x2, y2, z2,E,G,Iy,Iz,J,A):

    '''
    Return the element's local stiffness matrix (in global coordinates).

    
    Inputs:
        eleTag: int
            The tag of the element

        i_nodeTag: int
            The tag of i-node

        j_nodeTag: int
            The tag of j-node

        x1,y1,z1,x2,y2,z2: float
            The nodal coordiantes of the element

        E,G,Iy,Iz,J,A: float
            The properties of the element.

    Return:
        this_beamcol.K():   ndarray of shape (12,12)
            The stiffness matrix of the element.

    '''

    #Create a beam-column
    this_beamcol = BeamCol(eleTag, i_nodeTag, j_nodeTag, x1, y1, z1, x2, y2, z2, E, G, Iy, Iz, J, A)

    #Get the global stiffness matrix of this element
    return this_beamcol.K()

@jit
def Ele_K_to_Global(BeamCol):
    '''
    Return the [row, col, values] of element's local stiffness matrix in the global stiffness matrix.
    This will later be used for constructing the global stiffness matrix, which will be stored in a sparse matrix.
    Inputs:
        BeamCol:BeamCol() object
    
    Return:
        data: ndarray of shape(144,)
            The values of local stiffness matrix
        
        indices: ndarray of shape (144,2)
            The corresponding indices in global stiffness matrix.
            The first column is the row number, the second coclumn is the column number
    '''
    data = jnp.ravel(BeamCol.K()) #local stiffness matrix, flatten
    i =BeamCol.i_nodeTag #i-node
    j =BeamCol.j_nodeTag #j-node
    indices_dof = jnp.hstack((jnp.linspace(i*6,i*6+5,6,dtype=int),jnp.linspace(j*6,j*6+5,6,dtype=int))) #indices represented the dofs of this beamcol
    rows,cols = jnp.meshgrid(indices_dof,indices_dof,indexing='ij')#rows and columns
    indices = jnp.vstack((rows.ravel(),cols.ravel())).T #indices in global stiffness matrix
    return data,indices
    
@jit
def Ele_Sens_K_Coord(BeamCol):
    '''
    Return the sensitivity of element's 
    local stiffness matrix (in global coordinates) 
    w.r.t. the element's coordinates.

    
    Inputs:
        BeamCol: BeamCol() object

    Return:
        Sens_K_Coord: ndarray of shape (12,12,6)
            The derivatives of the stiffness matrix wrt to the nodal coordinates

    '''

    
    #Properties of this beam column
    eleTag = BeamCol.eleTag
    i_nodeTag, j_nodeTag = [BeamCol.i_nodeTag,BeamCol.j_nodeTag]
    x1,y1,z1,x2,y2,z2 = [BeamCol.x1,BeamCol.y1,BeamCol.z1,BeamCol.x2,BeamCol.y2,BeamCol.z2]
    E = BeamCol.E
    G = BeamCol.G
    Iy = BeamCol.Iy 
    Iz = BeamCol.Iz 
    J = BeamCol.J
    A = BeamCol.A
    
    #Calculate the sensitivity
    #argnums indicates the variables to which the Jacobian is calculated
    return jacfwd(BeamCol_K,argnums=(3,4,5,6,7,8))(eleTag, i_nodeTag, j_nodeTag, x1, y1, z1, x2, y2, z2, E, G, Iy, Iz, J, A)



"""
A optimization module for JAX-SSO.
Available optimizers:
    1. Vanilla gradient descent for unconstrained problems
    2. SLSQP for constrained problems from Nlopt (Install Nlopt: pip install nlopt) [https://nlopt.readthedocs.io/en/latest/NLopt_Python_Reference]
"""


import numpy as np
import nlopt

class Optimization():
    """
    A class for optimization
    """

    def __init__(self,n_x,method='GD'):
        """
        Initialize an Optimization() object.
        
        Parameters:
            method: str
                'GD': vanilla gradient descent, default
                'SLSQP': sequential least-squares quadratic programming:
                    Reference:Dieter Kraft, "A software package for sequential quadratic programming", Technical Report DFVLR-FB 88-28, Institut für Dynamik der Flugsysteme, Oberpfaffenhofen, July 1988.
               
            n_x: int
                number of parameters to optimize        
        """
        
        self.method = method      # Optimizer
        self.n_x = n_x   # Number of parameters

        self.if_eq_c = False # if there is equality constraint
        self.if_ineq_c = False #if there is inequality constraint
        self.if_bounds = False #if there are bounds for parameters
        self.if_maxiter = False # if there is stop criteria: Number of maximum function evaluation
        self.if_ftol_rel = False # if there is stop criteria: Relative tolerance
        self.if_ftol_abs = False # if there is stop criteria: Absolute tolerance
        self.GD_norm = False #if the gradients are normalized

    def set_objective(self,fun):
        """
        The objective function to minimize, and its gradient

        Parameters:
        -----
        fun: callable 
            fun(x, *args) -> tuple, (fun_value,grad_fun)
                fun_value: float
                    the objective value
        
                grad_fun: ndarry of shape (n_x,)
                    the gradient of the objective

        """   
        self.fun = fun #the objective function

    def set_maxiter(self,maxiter):
        """
        Set the maximum number of iterations

        Parameters:
        -----
        maxiter: int 
        """
        self.if_maxiter = True
        self.maxiter = maxiter #set maxiter

    def set_step_size(self,ln_rate):
        """
        Set the step size for gradient descent

        Parameters:
        -----
        ln_rate: float
        """
        
        if self.method == 'GD':
            self.ln_rate = ln_rate

    def set_GD_normalized(self,normalized=False):
        """
        If set to 'True', the gradient will be normalized by the max(abs) of the gradient.

        Parameters:
        -----
        ln_rate: float
        """
        
        if normalized:
            self.GD_norm = True
        else:
            self.GD_norm = False
    
    def set_ftol_rel(self,ftol_rel):
        """
        Stop criteria: relative function value

        Parameters:
        -----
        ftol_rel: float
        """
        self.if_ftol_rel = True
        self.ftol_rel = ftol_rel

    def set_ftol_abs(self,ftol_abs):
        """
        Stop criteria: relative function value

        Parameters:
        -----
        ftol_abs: float
        """
        self.if_ftol_abs = True
        self.ftol_abs = ftol_abs

    def set_bounds(self,x_lb,x_ub):
        """
        Setting bounds for parameters

        Parameters:
        -----
        x_lb: ndarray
            lower bounds
        
        x_ub: ndarray
            upper bounds
        """
        if self.method == 'GD':
            raise TypeError(" Gradient descent cannot deal with bounds, change \'method\'")
        else:
            self.if_bounds = True
            self.lb = x_lb
            self.ub = x_ub


    def set_inequality_constraints(self,fc,grad_fc,d_fc,tol=0):
        """
        Setting inequality constraints

        Parameters:
        -----
        fc: callable 
            fc(x, *args) -> float/ndarray of shape (d_fc,); function for constraints; fc(x) <= 0

        grad_fc: callable 
            grad_fc(x, *args) -> ndarray of shape (d_fc,n_x); gradients of constraints
        
        d_fc: int
            shape of the return of fc(x, *args); dimension of the constraints

        tol:float/ndarray of shape (d_fc,)
            tolerence for constraints; default is zero

        """
        if self.method == 'GD':
            raise TypeError(" Gradient descent cannot deal with constraints, change \'method\'")
        
        else:
            self.if_ineq_c = True
            self.ineq_fc = fc
            self.ineq_grad_fc = grad_fc
            self.d_ineq_c = d_fc
            if self.d_ineq_c > 1:
                if np.size(tol) == 1:
                    self.ineq_tol = tol * np.ones(self.d_ineq_c) 
                else:
                    self.ineq_tol = tol
            elif self.d_ineq_c == 1:
                self.ineq_tol = tol

    def set_equality_constraints(self,fc,grad_fc,d_fc,tol=0):
        """
        Setting equality constraints

        Parameters:
        -----
        fc: callable 
            fc(x, *args) -> float/ndarray of shape (d_fc,); function for constraints; fc(x) = 0

        grad_fc: callable 
            grad_fc(x, *args) -> ndarray of shape (d_fc,n_x); gradients of constraints
        
        d_fc: int
            shape of the return of fc(x, *args); dimension of the constraints

        tol:float/ndarray of shape (d_fc,)
            tolerence for constraints; default is zero

        """
        if method == 'GD':
            raise TypeError(" Gradient descent cannot deal with constraints, change \'method\'")
        else:
            self.if_eq_c = True 
            self.eq_fc = fc
            self.eq_grad_fc = grad_fc
            self.d_eq_c = d_fc
            if self.d_eq_c > 1:
                if np.size(tol) == 1:
                    self.eq_tol = tol * np.ones(self.d_eq_c) 
                else:
                    self.eq_tol = tol
            elif self.d_eq_c == 1:
                self.eq_tol = tol

    def Nlopt_objective(self):
        '''
        The objective function for Nlopt package.
        
        --------------------
        Form required by Nlopt
        --------------------
        def f(x, grad):
            if grad.size > 0:
               ...set grad to gradient, in-place...
               return ...value of f(x)...
        --------------------
        '''

        def f(x,grad):
            res_1,res_2 = self.fun(x)
            if grad.size>0:
                grad[:] = res_2

            return res_1


        return f

    def Nlopt_eq_constraints(self):
        '''
        The equality constraints for Nlopt package. Single-valued.
        
        --------------------
        Form required by Nlopt
        --------------------
        def f(x, grad):
            if grad.size > 0:
               ...set grad to gradient, in-place...
               return ...value of f(x)...
        --------------------
        '''

        #function for Nlopt
        def f(x,grad):
            if grad.size>0:
                grad[:] = self.eq_grad_fc(x)
            return self.eq_fc(x)

        return f

    def Nlopt_eq_mconstraints(self):
        '''
        The equality constraints for Nlopt package. Vector-valued.
        
        --------------------
        Form required by Nlopt
        --------------------
        def f(x, grad):
            if grad.size > 0:
               ...set grad to gradient, in-place...
               return ...value of f(x)...
        --------------------
        '''

        #function for Nlopt
        def c(result,x,grad):
            if grad.size>0:
                grad[:] = self.eq_grad_fc(x)
                result[:] = self.eq_fc(x)
            

        return c

    def Nlopt_in_constraints(self):
        '''
        The inequality constraints for Nlopt package.Single value.
        
        --------------------
        Form required by Nlopt
        --------------------
        def f(x, grad):
            if grad.size > 0:
               ...set grad to gradient, in-place...
               return ...value of f(x)...
        --------------------
        '''

        #function for Nlopt
        def f(x,grad):
            if grad.size>0:
                grad[:] = self.ineq_grad_fc(x)
            return self.ineq_fc(x)

        return f

    def Nlopt_in_mconstraints(self):
        '''
        The inequality constraints for Nlopt package. Vector-valued.
        
        --------------------
        Form required by Nlopt
        --------------------
        def f(x, grad):
            if grad.size > 0:
               ...set grad to gradient, in-place...
               return ...value of f(x)...
        --------------------
        '''

        #function for Nlopt
        def c(result,x,grad):
            if grad.size>0:
                grad[:] = self.ineq_grad_fc(x)
                result[:] = self.ineq_fc(x)
            

        return c

    #Conduct optimization
    def optimize(self,x_ini,log=True):
        '''
        Conduct optimization, given initial guess.
        
        Parameters:
        -----
        x_ini: ndarray
            initial guess of optimization parameters, shape of (n_x,)

        log: bool
            print log at each functional call

        Return:
        -----
        x_fin: ndarray
            optimized parameters
        --------------------
        '''

        #Vanilla gradient descent with/without normalized gradients
        if self.method == 'GD':
            i = 0 #initial iteration number
            current_x = x_ini #initial parameters
            f_store = [] # store the functional value
            while i < self.maxiter:
                obj,grad_obj = self.fun(current_x) #functional value and its gradient
                f_store.append (obj) #store the functional value
                normalized_grad = grad_obj #original gradient
                if self.GD_norm == True:
                    normalized_grad = normalized_grad/np.max(np.abs(normalized_grad)) #normalize the gradients
                if i>=1:
                    if self.if_ftol_rel==True:
                        if ((f_store[i-1] - f_store[i])/f_store[i-1]) <= self.ftol_rel[0]:
                            print('Stopping criteria \'ftol_rel\' met, end of optimization')
                            
                            break
                    if self.if_ftol_abs==True:
                        if f_store[i] <= self.ftol_abs[0]:
                            print('Stopping criteria \'ftol_abs\' met, end of optimization')
                            break

                
                current_x = current_x - self.ln_rate * normalized_grad #update parameters
                if log ==True:
                    print('Step {}, objective = {}'.format(i,f_store[i]))
                i += 1 #update iteration number

            x_fin = current_x #store the final value
            
            if i==self.maxiter:
                print('Stopping criteria \'maxiter\' met, end of optimization')

        #SLSQP, a SQP method
        elif self.method == 'SLSQP':
            opt = nlopt.opt(nlopt.LD_SLSQP,self.n_x) # opt objects

            #The objective function

            opt.set_min_objective(self.Nlopt_objective()) #objective function
            
            #bounds
            if self.if_bounds == True:
                opt.set_lower_bounds(self.lb)
                opt.set_upper_bounds(self.ub)

            #equality constraints
            if self.if_eq_c == True:
                if self.d_eq_c == 0:
                    opt.add_equality_constraint(self.Nlopt_eq_constraints(),self.eq_tol)
                else:
                    opt.add_equality_mconstraint(self.Nlopt_eq_mconstraints(),self.eq_tol)

            #inequality constraints
            if self.if_ineq_c == True:
                if self.d_ineq_c == 0:
                    opt.add_inequality_constraint(self.Nlopt_in_constraints(),self.ineq_tol)
                else:
                    opt.add_inequality_mconstraint(self.Nlopt_in_mconstraints(),self.ineq_tol)

            #stopping criteria
            if self.if_maxiter:
                opt.set_maxeval(self.maxiter)
            if self.if_ftol_abs:
                opt.set_ftol_abs(self.set_ftol_abs) 
            if self.if_ftol_rel:
                opt.set_ftol_rel(self.set_ftol_rel)

            #perform optimization
            x_fin = opt.optimize(x_ini)

        return x_fin

                







import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="JaxSSO",
    version="0.0.4",
    author="Gaoyuan Wu",
    author_email="gaoyuanw@princeton.edu",
    description="A framework for structural shape optimization based on automatic differentiation (AD) and the adjoint method, enabled by JAX",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/GaoyuanWu/JaxSSO",
    packages=setuptools.find_packages(include=['JaxSSO']),
    keywords=["jax", "automatic-differentiation", "shape optimization", "form-finding", "structural optimization"],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    install_requires=[
        'numpy>=1.22.0',
        'matplotlib',
        'jax'
    ],
    extras_require = {
        'FEA Solver': ['PyNiteFEA']
    },
    include_package_data = True,
    python_requires=">=3.7",
)



"""
References
---------
1. PyNite: An open-source FEA solver for python; https://github.com/JWock82/PyNite
2. Bathe, K. J. (2006). Finite element procedures. Klaus-Jurgen Bathe.
"""      
class Node():
    """
    Create a JaxSSO node.

    Input
    -----
    nodeTag: int 
        the tag/index of this node
    X, Y, Z: float
        the coordinates of thie node

    """
    
    def __init__(self, nodeTag, X, Y, Z):
        self.nodeTag = nodeTag      # Index of this node
        self.X = X            # Global X coordinate
        self.Y = Y            # Global Y coordinate
        self.Z = Z            # Global Z coordinate
        
